/* Copyright (c) 2000-2001 ActiveState Tool Corporation.
   See the file LICENSE.txt for licensing information. */
   
// Globals
var globalPrefService;//defined in koGlobalOverlay.js
var globalPrefs = null;

var log = getLoggingMgr().getLogger('filepicker.remote');

const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

const nsILocalFile = Components.interfaces.nsILocalFile;
const nsLocalFile_CONTRACTID = "@mozilla.org/file/local;1";
const nsIFile = Components.interfaces.nsIFile;

const koIFTPNode = Components.interfaces.koIFTPNode;
const koIFTPNodeFile = Components.interfaces.koIFTPNodeFile;
const koFTPNodeFile_CONTRACTID = "@activestate.com/koFTPNodeFile;1";

const nsIFilePicker     = Components.interfaces.nsIFilePicker;
const modeGetFolder = nsIFilePicker.modeGetFolder;
const modeOpen      = nsIFilePicker.modeOpen;
const modeSave      = nsIFilePicker.modeSave;
const returnCancel      = nsIFilePicker.returnCancel;
const returnOK      = nsIFilePicker.returnOK;
const returnReplace      = nsIFilePicker.returnReplace;

const nsIDirectoryServiceProvider = Components.interfaces.nsIDirectoryServiceProvider;
const nsIDirectoryServiceProvider_CONTRACTID = "@mozilla.org/file/directory_service;1";
const nsITreeBoxObject = Components.interfaces.nsITreeBoxObject;

// for now, we're only doing ftp browsing with this file browser
var FTPBrowserOnly = true;
const localDrives = "Local Drives";
var _fileListSep = "; ";
var sfile = null;
var retvals;
var filePickerMode;
var homeDir;
var treeView;
var helpTag;

var textInput;
var okButton;

function localFile(path) {
    this._sfile = Components.classes[nsLocalFile_CONTRACTID].createInstance(nsILocalFile);
    this._haveChildren = 0;
    this._directories = [];
    this._filelist = [];
    this._parent = null;
    if (path) this.init(path);
}

localFile.prototype = {
    get URL() { 
        return this._sfile.URL;
    },

    get path() { return this._sfile.path; },

    get parent() {
        if (!this._parent && this._sfile.parent) {
            this._parent = new localFile();
            this._parent._sfile = this._sfile.parent;
        }
        return this._parent;
    },

    get isDirectory() { return this._sfile.exists() && this._sfile.isDirectory(); },

    get isFile() { return this._sfile.exists() && this._sfile.isFile(); },

    get isSymlink() { return this._sfile.exists() && this._sfile.isSymlink(); },

    get fileSize() { return this._sfile.fileSize; },

    get lastModifiedTime() {return this._sfile.lastModifiedTime; },

    get target() { return this._sfile.target; },

    get leafName() { return this._sfile.leafName; },
    
    set followLinks(v) { this._sfile.followLinks = v; },
    get followLinks() { return this._sfile.followLinks; },
    
    get directories() {
        if (!this._haveChildren) this.initChildren();
        return this._directories;
    },
    get filelist() {
        if (!this._haveChildren) this.initChildren();
        return this._filelist;
    },
    createDirectory: function(dirName) {
        // what are the correct permissions to create a directory with?
        if (!this.isDirectory) return false;
        try {
            log.debug("createDirectory: current dir: "+this._sfile.path+", making "+dirName);
            // XXX funky nsIFile behaviour, you must append the new directory name
            // first, then call createUnique with the same directory name again!
            // I've duplicated this odd behaviour in koFTP.
            file = this.clone();
            file._sfile.append(dirName);
            file._sfile.createUnique(file._sfile.DIRECTORY_TYPE,755);
        } catch (e) {
            dialog_alert("Unable to create directory. "+this.lastError());
            return false;
        }
        return true;
    },
    remove: function(recursive) {
        try {
            this._sfile.remove(recursive);
            return true;
        } catch (e) {
            dialog_alert("Unable to remove "+this.path+"\n "+this.lastError());
            return false;
        }
    },
    initChildren: function() {
        this._directories = [];
        this._filelist = [];
        var dirEntries = this._sfile.QueryInterface(nsIFile).directoryEntries;
        var nextFile = null;
        while (dirEntries.hasMoreElements()) {
          nextFile = dirEntries.getNext().QueryInterface(nsIFile);
          // XXXjag hack for bug 82355 till symlink handling is fixed (bug 57995)
          var isDir = false;
          try {
            isDir = nextFile.isDirectory();
          } catch (e) {
            // this here to fool the rest of the code into thinking
            // this is a nsIFile object
            log.debug("initChildren: nextfile : "+nextFile.leafName);
            nextFile = { leafName : nextFile.leafName,
                         fileSize : 0,
                         lastModifiedTime : 0,
                         isHidden: function() { return false; } };
          }
          // end of hack
          fileobj = new Object();
          fileobj.file = new localFile();
          fileobj.file._parent = this;
          fileobj.file._sfile = nextFile;
          if (isDir) {
            this._directories[this._directories.length] = fileobj;
          } else {
            this._filelist[this._filelist.length] = fileobj;
          }
        }
        this._haveChildren = 1;
    },
    
    init: function(path) {
        this._sfile.initWithPath(path);
        this._sfile.normalize();
    },
    isHidden: function() {
        return this._sfile.isHidden();
    },
    exists: function() {
        return this._sfile.exists();
    },
    clone: function() {
        var newfile = new localFile();
        newfile._sfile = this._sfile.clone().QueryInterface(nsILocalFile);
        newfile._parent = this._parent;
        newfile._directories = this._directories;
        newfile._filelist = this._filelist;
        return newfile;
    },
    equals: function(file) {
        return this._sfile.equals(file._sfile);
    },
    getChild: function(URL) {
        var i;
        for (i = 0; i < this._directories.length; i++) {
            if (this._directories[i].file.URL == URL) return this._directories[i].file;
        }
        for (i = 0; i < this._filelist.length; i++) {
            if (this._filelist[i].file.URL == URL) return this._filelist[i].file;
        }
        return null;
    },
    appendRelativePath: function(input) {
        this._sfile.appendRelativePath(input);
    },
    moveTo: function(parent, newname) {
        try {
            this._sfile.moveTo(parent, newname);
        } catch (e) {
            dialog_alert("Unable to rename file. "+this.lastError());
        }
    },
    disconnect: function() {
    },
    refresh: function() {
    },
    lastError: function() {return "";}
}


function FTPFile(path, server, port, username, password) {
    this._haveChildren = 0;
    this._directories = [];
    this._filelist = [];
    // allow us to create an empty ftpfile
    this._sfile = Components.classes[koFTPNodeFile_CONTRACTID].createInstance(koIFTPNodeFile);
    if (typeof(path)=="undefined") return;
    
    log.debug("FTPFile.constructor("+path+", "+server+", "+port+", "+username+", "+password+")");
    if (typeof(port) != "number") port = Number(port);
    if (server != "") {
        try {
            this._sfile.init(server, port, username, password);
        } catch (e) {
            dialog_alert("Unable to connect to "+server);
            //log.exception(e);
            return;
        }
    }
    if (path) this.init(path);
}

// The following two lines ensure proper inheritance (see Flanagan, p. 144).
FTPFile.prototype = new localFile();
FTPFile.prototype.constructor = FTPFile;

FTPFile.prototype.__defineGetter__("parent",
function() 
{
    if (!this._parent && this._sfile.parent) {
        this._parent = new FTPFile();
        this._parent._sfile = this._sfile.parent;
    }
    return this._parent;
});

// our koFTPFile now returns seconds rather than
// milliseconds, which is contrary to nsIFile idl, but
// eases use in the rest of komodo which is written to
// use seconds.  However, our treeview expects milliseconds
// so we make the tranformation here.
FTPFile.prototype.__defineGetter__("lastModifiedTime",
function() 
{
    return this._sfile.lastModifiedTime*1000;
});

FTPFile.prototype.initChildren = function() {
    this._directories = [];
    this._filelist = [];
    var dirEntries = this._sfile.QueryInterface(koIFTPNode).directoryEntries;
    var nextFile = null;
    while (dirEntries.hasMoreElements()) {
      nextFile = dirEntries.getNext().QueryInterface(koIFTPNode);
      // XXXjag hack for bug 82355 till symlink handling is fixed (bug 57995)
      var isDir = false;
      try {
        isDir = nextFile.isDirectory();
      } catch (e) {
        // this here to fool the rest of the code into thinking
        // this is a nsIFile object
        log.debug("FTPFile.initChildren: nextfile 2: "+nextFile.leafName);
        nextFile = { leafName : nextFile.leafName,
                     fileSize : 0,
                     lastModifiedTime : 0,
                     isHidden: function() { return false; } };
      }
      // end of hack
      log.debug("FTPFile.initChildren: child is "+nextFile.path);
      var fileobj = new Object();
      fileobj.file = new FTPFile();
      fileobj.file._parent = this;
      fileobj.file._sfile = nextFile;
      log.debug("fileobj : "+nextFile.leafName);
      fileobj.cachedName = nextFile.leafName;
      fileobj.cachedDate = fileobj.file.lastModifiedTime;
      fileobj.cachedDateText = formatDate(fileobj.cachedDate);
      fileobj.cachedSize = String(nextFile.fileSize);
      if (isDir) {
        this._directories[this._directories.length] = fileobj;
      } else {
        this._filelist[this._filelist.length] = fileobj;
      }
    }
    this._haveChildren = 1;
}
FTPFile.prototype.initWithConnection = function(conn, path) {
    this._sfile.initWithConnection(conn, path);
}
FTPFile.prototype.init = function(path) {
    this._sfile.initWithPath(path);
    log.debug("FTPFile.init: path is "+this._sfile.path);
}
FTPFile.prototype.clone = function() {
    var newfile = new FTPFile();
    newfile._sfile = this._sfile.clone().QueryInterface(koIFTPNodeFile);
    newfile._parent = this._parent;
    newfile._directories = this._directories;
    newfile._filelist = this._filelist;
    return newfile;
}
FTPFile.prototype.disconnect = function() {
    this._sfile.connection.close();
}
FTPFile.prototype.refresh = function() {
    log.debug("FTPFile.refresh "+this._sfile.path);
    this._sfile.connection.refresh(this._sfile.path);
}
FTPFile.prototype.lastError = function() {return this._sfile.connection.lastError;}

function filebrowserLoad() {
  globalPrefService = Components.classes["@activestate.com/koPrefService;1"].
    getService(Components.interfaces.koIPrefService);
  globalPrefs = globalPrefService.prefs;

  textInput = document.getElementById("textInput");
  okButton = document.getElementById("ok");
  treeView = new filebrowseview();
  treeView.selectionCallback = onSelect;

  if (window.arguments) {
    var o = window.arguments[0];
    retvals = o.retvals; /* set this to a global var so we can set return values */
    const title = o.title;
    filePickerMode = o.mode;
    if (o.displayDirectory) {
      const directory = o.displayDirectory.path;
    }
    const initialText = o.defaultString;
    const filterTitles = o.filters.titles;
    const filterTypes = o.filters.types;
    const numFilters = filterTitles.length;
    helpTag = o.helpTag;

    window.title = title;

    if (initialText) {
      textInput.value = initialText;
    }
  }

  if ((filePickerMode == modeOpen) ||
      (filePickerMode == modeSave)) {

    if (filePickerMode == modeSave)
        treeView.singleSelect = true;

    treeView.setFilter(filterTypes[0]);

    /* build filter popup */
    var filterPopup = document.getElementById("filterMenu");//document.createElement("menupopup");

    for (var i = 0; i < numFilters; i++) {
      var menuItem = document.createElement("menuitem");
      menuItem.setAttribute("label", filterTitles[i] + " (" + filterTypes[i] + ")");
      menuItem.setAttribute("filters", filterTypes[i]);
      filterPopup.appendChild(menuItem);
    }

    var filterMenuList = document.getElementById("filterMenuList");
    //filterMenuList.appendChild(filterPopup);
    if (numFilters > 0)
      filterMenuList.selectedIndex = 0;
    var filterBox = document.getElementById("filterBox");
    filterBox.removeAttribute("hidden");
  } else if (filePickerMode == modeGetFolder) {
    treeView.singleSelect = true;
    treeView.showOnlyDirectories = true;
  }

  // start out with a filename sort
  handleColumnClick("FilenameColumn");

  try {
    var buttonLabel = getOKAction();
    okButton.setAttribute("label", buttonLabel);
  } catch (exception) {
    // keep it set to "OK"
  }

  // setup the dialogOverlay.xul button handlers
  doSetOKCancel(onOK, onCancel);
  retvals.buttonStatus = returnCancel;

  var tree = document.getElementById("directoryTree");
  tree.treeBoxObject.view = treeView;

  // Start out with the ok button disabled since nothing will be
  // selected and nothing will be in the text field.
  okButton.disabled = true;
  textInput.focus();

  // load the deployment server list
  if (directory.indexOf("ftp://") != 0) {
      log.debug("filebrowserLoad: loading server list");
      loadServerList();
  }
  
  // This allows the window to show onscreen before we begin
  // loading the file list
  setTimeout(setInitialDirectory, 0, directory);
}

function filebrowserClose() {
    closeFTP();
}

function getURIInfo(uri) {
    var uriparts = new Array();
    uriparts["host"]="";
    uriparts["port"]="";
    uriparts["username"]="";
    uriparts["password"]="";
    uriparts["path"]="";
    var user = "";
    var host = "";
    var site=null;

    if (uri.indexOf("ftp://") != 0) {
        log.error("FTP getURIInfo error, receiving a non ftp url: "+uri);
        return null;
    }
    uri = uri.substr(6); // remove 'ftp://'
    // now have: userinfo@siteinfo/pathinfo
    var delim = uri.indexOf("/");
    if (delim > 0) {
        site = uri.substr(0,delim);
        if (delim < uri.length)
            uriparts["path"] = uri.substr(delim);
    } else {
        site=uri;
    }
    log.debug("getURIInfo site : "+site);
    log.debug("getURIInfo path : "+uriparts["path"]);
        
    // we've received a string that can be contain any combination of parts of the following,
    // the only absolute is the host
    // user:email@host@host.com:port
    delim = site.lastIndexOf("@");
    // we have 1-3 elements
    if (delim < 0) {
        host = site;
    } else {
        user = site.substr(0,delim);
        host = site.substr(delim+1);
    }
    
    // create host and port
    var hparts = host.split(":");
    uriparts["host"]=hparts[0];
    if (hparts.length > 1) uriparts["port"] = hparts[1];

    delim = user.indexOf(":");
    if (delim < 0) {
        uriparts["username"] = user;
    } else {
        uriparts["username"] = user.substr(0,delim);
        uriparts["password"] = user.substr(delim+1);
    }
    return uriparts;
}

function FTPInfo() {
    this.username = null;
    this.password = null;
    this.host = null;
    this.port = null;
    this.connection = null;
}
FTPInfo.prototype.constructor = FTPInfo;

var currentServer = null;
var ftpBrowser = false;
var ftpInfo = new FTPInfo();
var lastLocalDirectory = null;
function newFile(directory)
{
    log.debug("newFile: creating new file for "+directory);
    var file = null;
        
    if (directory.indexOf("ftp://") == 0) {
        var parts = getURIInfo(directory);
        
        var path = parts["path"];
        if (path == "") {
            if (!ftpInfo.connection) {
                // first connection, go home
                path = "~";
            } else {
                path = "/";
            }
        }
        if (path != "~" && path[0] != "/") path = "/" + path;
        
        ftpInfo.username = parts["username"];
        ftpInfo.password = parts["password"];
        ftpInfo.domain = parts["host"];
        ftpInfo.port = parts["port"];
        ftpBrowser = true; //once ftp, always ftp!
        log.debug("newFile: u: "+ftpInfo.username+" p: "+ftpInfo.password+" d: "+ftpInfo.domain+":"+ftpInfo.port);
        // ask for username/password if it isn't provided
        if (ftpInfo.username == "" || ftpInfo.password == "") {
            // XXX fix persist post-1.2
            var obj = dialog_authenticate('Authenticate',
                                 'Please enter your username and password',
                                 ftpInfo.username, true, false);
            if (!obj) {
                log.debug('login canceled by authentication dialog');
                return null;
            }
            if (obj.retval != "Cancel") {
                ftpInfo.username = obj.username;
                ftpInfo.password = obj.password;
                if (obj.anonymous)  {
                    log.debug('login will be anonymous');
                    // attempt anonymous
                    if (!ftpInfo.username) ftpInfo.username = 'anonymous';
                    if (!ftpInfo.password) ftpInfo.password = 'anon@anon.org';
                }
            }
            else  {
                log.debug('login canceled by authentication dialog');
                return null;
            }
            log.debug("newFile: u: "+ftpInfo.username+" p: "+ftpInfo.password+" d: "+ftpInfo.domain+":"+ftpInfo.port);
        }

        //  pass exceptions, connection will have error message.
        try {
            log.debug('newFile: create FTPFile for '+path);
            if (path == "~" || path == "/~") {
                file = new FTPFile("", ftpInfo.domain, ftpInfo.port, ftpInfo.username, ftpInfo.password);
                var home = file._sfile.connection.getHomeDirectory();
                if (home) file.init(home);
            } else {
                file = new FTPFile("", ftpInfo.domain, ftpInfo.port, ftpInfo.username, ftpInfo.password);
                file.init(path);
            }
        } catch(e) {
            /* do nothing, we deal with the error later */
            //log.exception(e,'newFile: failed to create FTPFile');
        }
        
        ftpInfo.connection = file._sfile.connection;
        if (ftpInfo.connection.lastError != "") {
            dialog_alert(ftpInfo.connection.lastError);
            file = null;
            return null;
        }
        return file;
    }
    if (ftpBrowser) {
        // maintain a single ftp connection by passing our connection around
        log.debug('newFile: create FTPFile using existing connection '+directory);
        file = new FTPFile();
        file.initWithConnection(ftpInfo.connection, directory);
        return file;
    }
    if (FTPBrowserOnly) {
        return null;
    }
    log.debug('newFile: set lastLocalDirectory '+directory);
    ftpInfo.connection = null;
    lastLocalDirectory = directory;
    return new localFile(directory)
}

function closeFTP()
{
    if (ftpBrowser) {
        ftpBrowser = false;
        if (ftpInfo.connection) ftpInfo.connection.close();
        ftpInfo = new FTPInfo();
        ftpInfo.connection = null;
    }
}

function setInitialDirectory(directory)
{
  // get the home dir
  log.debug("setInitialDirectory: ["+directory+"]");
  clearHistoryList();
  var dirServiceProvider = Components.classes[nsIDirectoryServiceProvider_CONTRACTID]
                                     .getService(nsIDirectoryServiceProvider);
  var persistent = new Object();
  homeDir = dirServiceProvider.getFile("Home", persistent);
  sfile = null;
  if (directory) {
    sfile = newFile(directory);
    if (sfile)
        log.debug("setInitialDirectory: file: "+sfile.path+" exists: "+sfile.exists()+" isDir "+sfile.isDirectory);
    else
        log.debug("setInitialDirectory: no file produced for "+directory);
  } else if (!FTPBrowserOnly && !(sfile.exists() && sfile.isDirectory)) {
    // Start in the user's home directory
    log.debug("setInitialDirectory: initialize new file with homeDir "+homeDir.path);
    sfile = newFile(homeDir.path);
  }
  if (!sfile) {
      treeView.clearList();
      return;
  }
  gotoDirectory(sfile);
}

function onFilterChanged(target)
{
  // Do this on a timeout callback so the filter list can roll up
  // and we don't keep the mouse grabbed while we are refiltering.

  setTimeout(changeFilter, 0, target.getAttribute("filters"));
}

function changeFilter(filterTypes)
{
  window.setCursor("wait");
  treeView.setFilter(filterTypes);
  window.setCursor("auto");
}

function cloneAppend(input) {
      var file = sfile.clone();
      if (!file) {
        log.debug("unable to clone!");
        return null;
      }
      dump('input is '+input+'\n');
      /* XXX we need an XP way to test for an absolute path! */
      if (input[0] == '/' || input[1]==":")   /* an absolute path was entered */
        file.init(input);
      else if ((input.indexOf("/../") > 0) ||
               (input.substr(-3) == "/..") ||
               (input.substr(0,3) == "../") ||
               (input == "..")) {
        /* appendRelativePath doesn't allow .. */
        file.init(file.path + "/" + input);
      } else {
        try {
          file.appendRelativePath(input);
        } catch (e) {
          log.exception(e,"Can't append relative path '"+input+"'");
          return null;
        }
      }
      if (!file.exists()) {
          // check to see if this is a directory since python ftplib won't let us know
          var child = sfile.getChild(file.URL);
          if (child) file = child;
      }
      return file;
}

function onOK()
{
  var errorTitle, errorMessage, promptService;
  var ret = returnCancel;
  var file;
  var i=0;
  // get the files from the tree
  var files = treeView.getSelectedFiles();
  var input = textInput.value;
  var inputFiles = input.split(_fileListSep);
  log.debug("onOK: number of files: "+files.length);
  log.debug("onOK: input is: "+input);
  log.debug("onOK: first input "+inputFiles[0]);
  log.debug("onOK: using basedir from: "+sfile.URL);

    if (filePickerMode == modeSave) {
      // dump any select files since we have an input filename
      files = new Array();
    }

  // we don't have files in the file picker, they typed something in, create
  // the files array
  if (inputFiles.length < 1) {
      dump ("no inputFiles!\n");
      return false;
  }
  
  for (i = 0; i < inputFiles.length; i++) {
      log.debug("checking input "+inputFiles[i]);
      input = inputFiles[i];
      if (input[0] == '~') { // XXX XP?
        if (ftpBrowser) {
            var home = sfile._sfile.connection.getHomeDirectory();
            input = home + input.substring(1);
            log.debug("FTP HOME: "+input);
        } else {
            input  = homeDir.path + input.substring(1);
        }
      }
      
      // look to see if we already have this file
      var haveFile = false;
      for (i = 0; i< files.length; i++) {
          var url = sfile.URL;
          
          // make sure we have a slash!
          if (url[url.length-1] != "/" && input[0] != "/") url += "/"+ input;
          else url += input;
          
          log.debug("checking "+files[i].URL+" == "+url);
          if (files[i].URL == url) {
              haveFile = true;
              break;
          }
      }
      if (haveFile) continue;
      file = cloneAppend(input);
      log.debug("adding file "+file.URL+" to list isDir:"+file.isDirectory);
      files.push(file);
  }

  // verify that the file(s) or directory exist
  for (i = 0; i < files.length; i++) {
      file = files[i];
      if (!file.exists() && (filePickerMode != modeSave)) {
        errorTitle = 'Error opening file';
        errorMessage = "File "+file.path+" doesn't exist";
        promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                                  .getService(Components.interfaces.nsIPromptService);
        promptService.alert(window, errorTitle, errorMessage);
        return false;
      }
  }
  
  // unless we are in modeOpen, we will not have more than one file
  file = files[0];
  var filepaths = new Array();
  retvals.filepaths = filepaths; // set an empty return array
  // return the server object so we don't have to look it up again!
  var idx = getAliasIndex(currentServer);
  if (idx > -1) retvals.server = servers[idx];
  
  switch(filePickerMode) {
  case modeOpen:
    if (file.isFile) {
        // if the first file is a file, all files 'should' be files
        // if the first file is a file, all files 'should' be files
        for (i=0; i < files.length; i++) {
            // build an array of full file paths
            log.debug("adding path "+files[i].URL);

            // munge the url into a koRemote url
            url = files[i].URL;
            if (url.substring(0, 7) != 'file://') {
                // retvals.server exists only for ftp aliases from preferences
                // typed in sites we skip that part
                if (retvals.server) {
                    url = "koremote://"+retvals.server.alias+"/"+url;
                } else {
                    url = "koremote:///"+url;
                }
            }
            log.debug("path added "+url);
            filepaths.push(url);
        }
        retvals.filepaths = filepaths;
        // single selection support
        retvals.directory = file.parent.path;
        ret = returnOK;
    } else if (file.isDirectory) {
      if (!sfile.equals(file)) {
        gotoDirectory(file);
      }
      textInput.value = "";
      doEnabling();
      ret = returnCancel;
    }
    break;
  case modeSave:
    if (file.isFile) { // can only be true if file.exists()
      // we need to pop up a dialog asking if you want to save
      var message = file.path+" already exists.\nDo you want to replace it?";
      var rv = window.confirm(message);
      if (rv) {
        ret = returnReplace;
        retvals.directory = file.parent.path;
      } else {
        ret = returnCancel;
      }
    } else if (file.isDirectory) {
      if (!sfile.equals(file)) {
        gotoDirectory(file);
      }
      textInput.value = "";
      doEnabling();
      ret = returnCancel;
    } else {
      var parent = file.parent;
      if (parent.exists() && parent.isDirectory) {
        ret = returnOK;
        retvals.directory = parent.path;
      } else {
        var oldParent = parent;
        while (!parent.exists() && parent != parent.parent) {
          oldParent = parent;
          parent = parent.parent;
        }
        errorTitle = "Error saving "+file.path;
        if (parent.isFile) {
          errorMessage = parent.path+" is a file, can't save "+file.path;
        } else {
          errorMessage = "Path "+oldParent.path+" doesn't exist, can't save "+file.path;
        }
        promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                                  .getService(Components.interfaces.nsIPromptService);
        promptService.alert(window, errorTitle, errorMessage);
        ret = returnCancel;
      }
    }
    break;
  case modeGetFolder:
    if (file.isDirectory) {
      retvals.directory = file.path;
    } else { // if nothing selected, the current directory will be fine
      retvals.directory = sfile.path;
    }
    ret = returnOK;
    break;
  }

  //retvals.file = file.path;
    url = file.URL;
    if (url.substring(0, 7) != 'file://') {
        // retvals.server exists only for ftp aliases from preferences
        // typed in sites we skip that part
        if (retvals.server) {
            url = "koremote://"+retvals.server.alias+"/"+url;
        } else {
            url = "koremote:///"+url;
        }
    }
  retvals.file = url;
  retvals.buttonStatus = ret;
  
  
  if (ret == returnCancel)
    return false;
  else {
    closeFTP(); // in case it's open
    return true;
  }
}

function onCancel()
{
  closeFTP(); // in case it's open
  // Close the window.
  retvals.buttonStatus = returnCancel;
  retvals.file = null;
  return true;
}

function onDblClick(e) {
  var t = e.originalTarget;
  if (t.localName != "treechildren")
    return;

  openSelectedFile();
}

function openSelectedFile() {
  var files = treeView.getSelectedFiles();
  if (!files.length)
    return;
  var hasDirectory = false;
  for (var i=0; i< files.length; i++) {
      var file = files[i];
      if (file.isDirectory) {
        gotoDirectory(file);
        hasDirectory = true;
        break;
      }
      if (file.isSymlink)
        file.init(file.target);
  }

  if (hasDirectory) return;

  // finish up!
  doOKButton();
}

function onClick(e) {
  var t = e.originalTarget;
  if (t.localName == "treecol")
    handleColumnClick(t.id);
}

function convertColumnIDtoSortType(columnID) {
  var sortKey;
  
  switch (columnID) {
  case "FilenameColumn":
    sortKey = filebrowseview.SORTTYPE_NAME;
    break;
  case "FileSizeColumn":
    sortKey = filebrowseview.SORTTYPE_SIZE;
    break;
  case "LastModifiedColumn":
    sortKey = filebrowseview.SORTTYPE_DATE;
    break;
  default:
    dump("unsupported sort column: " + columnID + "\n");
    sortKey = 0;
    break;
  }
  
  return sortKey;
}

function handleColumnClick(columnID) {
  var sortType = convertColumnIDtoSortType(columnID);
  var sortOrder = (treeView.sortType == sortType) ? !treeView.reverseSort : false;
  treeView.sort(sortType, sortOrder, false);
  
  // set the sort indicator on the column we are sorted by
  var sortedColumn = document.getElementById(columnID);
  if (treeView.reverseSort) {
    sortedColumn.setAttribute("sortDirection", "descending");
  } else {
    sortedColumn.setAttribute("sortDirection", "ascending");
  }
  
  // remove the sort indicator from the rest of the columns
  var currCol = document.getElementById("directoryTree").firstChild;
  while (currCol) {
    while (currCol && currCol.localName != "treecol")
      currCol = currCol.nextSibling;
    if (currCol) {
      if (currCol != sortedColumn) {
        currCol.removeAttribute("sortDirection");
      }
      currCol = currCol.nextSibling;
    }
  }
}

function doSort(sortType) {
  treeView.sort(sortType, false);
}

function onKeypress(e) {
  if (e.keyCode == 8) /* backspace */
    goUp();
  else if (e.keyCode == 13) { /* enter */
    var file = treeView.getSelectedFile();
    if (file) {
      if (file.isDirectory) {
        gotoDirectory(file);
        e.preventDefault();
      }
    }
  }
}

function doEnabling() {
  // Maybe add check if textInput.value would resolve to an existing
  // file or directory in .modeOpen. Too costly I think.
  var enable = (textInput.value != "");

  okButton.disabled = !enable;
}

function onTreeFocus(event) {
  // Reset the button label and enabled/disabled state.
  onSelect(treeView.getSelectedFiles());
}

function getOKAction(file) {
  if (file && file.isDirectory && filePickerMode != modeGetFolder)
    return 'Open';

  switch(filePickerMode) {
  case modeGetFolder:
    return 'Select';
    break;
  case modeOpen:
    return 'Open';
    break;
  case modeSave:
    return 'Save';
    break;
  }
  return ""; // js strict
}

function onSelect(files) {
    var filesText = "";
    var file = null;
    for (var i = 0; i < files.length; i++) {
        file = files[i];
        if (file) {
            var path = file.leafName;
            
            // break out on the first directory
            if (file.isDirectory) {
                if (filePickerMode == modeGetFolder) filesText = path;
                break;
            } else if (!(filePickerMode == modeGetFolder)) {
                if (path) {
                  if ((filePickerMode == modeGetFolder) || !file.isDirectory) {
                    if (filesText != "")
                        filesText += _fileListSep+path;
                    else
                        filesText = path;
                  }
                }
            }
        }
    }
    textInput.value = filesText;
    if (file && filesText != "") {
        var buttonLabel = getOKAction(file);
        okButton.setAttribute("label", buttonLabel);
        okButton.disabled = false;
        return;
    }

  okButton.disabled = true;
}

function onTextFieldFocus() {
  var buttonLabel = getOKAction(null);
  okButton.setAttribute("label", buttonLabel);
  doEnabling();
}

function onDirectoryChanged(target)
{
  var path = target.getAttribute("label");
  log.debug("onDirectoryChanged("+path+")");
  var file = newFile(path);

  if (!sfile.equals(file)) {
    // Do this on a timeout callback so the directory list can roll up
    // and we don't keep the mouse grabbed while we are loading.

    setTimeout(gotoDirectory, 0, file);
  }
}

function addToHistory(directoryName) {
  var menu = document.getElementById("lookInMenu");
  var menuList = document.getElementById("lookInMenuList");
  var children = menu.childNodes;

  if (!children) {
    var menuItem = document.createElement("menuitem");
    menuItem.setAttribute("label", directoryName);
    menu.appendChild(menuItem);
    menuList.selectedIndex = 0;
    return;
  }
  
  var i = 0;

  while (i < children.length) {
    if (children[i].getAttribute("label") == directoryName)
      break;

    ++i;
  }

  if (i < children.length) {
    if (i != 0) {
      var node = children[i];
      menu.removeChild(node);
      menu.insertBefore(node, children[0]);
    }
  } else {
    var menuItem = document.createElement("menuitem");
    menuItem.setAttribute("label", directoryName);
    menu.insertBefore(menuItem, children[0]);
  }

  menuList.selectedIndex = 0;
}

function goLast() {
  var menu = document.getElementById("lookInMenu");
  var children = menu.childNodes;
  var dir = null;
  if (children.length > 1) {
      dir = children[1];
      // move current history item to last history item
      var node = children[0];
      menu.appendChild(node, children[children.length-1]);
      onDirectoryChanged(dir);
  }
}

function doMKDIR() {
    if (!sfile) return false;
    var textInput = document.getElementById("textInput")
    var dirName = dialog_prompt("Enter the name for the new directory", null,textInput.value)
    if (!dirName) return false;
    
    log.debug("current dir "+sfile.path+" isdir: "+sfile.isDirectory);
    log.debug("doMKDIR: "+dirName);

    window.setCursor("wait");
    var e = new Object();
    var ret = false;
    if (sfile.createDirectory(dirName)) {
        log.debug("name of new folder is "+sfile.path+"/"+dirName);
        sfile.refresh();
        gotoDirectory(sfile);
        // select the new folder in the tree
        /* until we can edit cells in an ouliner, this does nothing
        if (treeView.selectEntryByNameType(dirName, 1)) {
            // make the cell editable, and edit it!
            doRename();
        }
        */
        ret = true;
    }
    window.setCursor("auto");
    return ret;
}

function doRename() {
    var textInput = document.getElementById("textInput")
    var newName = dialog_prompt("Enter the new name for the file or directory", null,textInput.value)
    if (!newName || newName == textInput.value) return false;

    file = treeView.getSelectedFile()
    log.debug("doRename called on "+file.path);
    file.moveTo(null, newName);
    
    sfile.refresh();
    gotoDirectory(sfile);
    return true;
}

function doDelete() {
    files = treeView.getSelectedFiles()
    if (dialog_yesNo("Are you sure you want to delete the selected files?", "No") == "No") {
      return;
    }
    for (var i=0; i< files.length; i++) {
        log.debug("doDelete called on "+files[i].path);
        var recurse = false;
        if (files[i].isDirectory) {
            if (dialog_yesNo("The file '"+files[i].leafName+"' is a Directory, would you like to delete all sub-directories and folders?", "No") == "Yes") {
              recurse = true
            }
        }
        files[i].remove(recurse);
        // XXX TODO remove from history if its in there
    }
    var tree = document.getElementById("directoryTree");
    tree.treeBoxObject.scrollToRow(0);
    sfile.refresh();
    gotoDirectory(sfile);
}

function doRefresh() {
    if (!sfile) return;
    window.setCursor("wait");
    sfile.refresh();
    gotoDirectory(sfile);
    window.setCursor("auto");
}

function goUp() {
  try {
    var parent = sfile.parent;
  } catch(ex) { dump("can't get parent directory\n"); }

  if (parent) {
    gotoDirectory(parent);
  }
}

function gotoDirectory(directory) {
    log.debug('gotoDirectory '+directory);
    addToHistory(directory.path);
  
    window.setCursor("wait");
    treeView.setDirectory(directory.path);
    window.setCursor("auto");
  
    sfile = directory;
}


var servers = null;
function Server(type, alias, hostname, port, path, username, password) {
  this.type = type;
  this.alias = alias;
  this.hostname = hostname;
  this.port = port;
  this.path = path;
  this.username = username;
  this.password = password;
  this.id = Server_generateId();
}

// Each Association object has a unique identifier.  This is used to
// establish a relationship between the association object and the
// items in the tree that represent it.

var Server_index = 0;  // used for generating ids
function Server_generateId() {
  return "server" + Server_index++;
}

/* XXX duplicated in pref-servers.js */
function decodePassword(password) {
    if (typeof(password) == 'undefined' || !password) return '';
    // XXX 2way encrypt the password somehow, to prevent plaintext in the prefs files
    var crypto = 
       Components.classes["@activestate.com/koPasswordCrypto;1"]
          .getService(Components.interfaces.koIPasswordCrypto);
    var text = password;
    try {
        text = crypto.decodePassword(password);    
    } catch(e) { 
        /* password is empty, that's ok, we'll ask for it later */ 
        log.warn("Dropping password that could not be decoded. See bug 18752.");
    }
    return text
}

function _strcmp(a,b) {
    if (a==b) return 0;
    if (a<b) return -1;
    return 1;
}

function getServerListFromPreference(prefset) {
  var list = new Array();  
  if (prefset.hasPref("deploymentServerList")) {
    var deploymentServerList = prefset.getPref("deploymentServerList");
    for (var i = 0; i < deploymentServerList.length; i++) {
      var tuple = deploymentServerList.getPref(i).QueryInterface(Components.interfaces.koIOrderedPreference);
      // invocations were added later
      var info = new String(tuple.getStringPref(1));
      var server_info = info.split(":")
      var password = decodePassword(server_info[5]);
      list.push(new Server(server_info[0], tuple.getStringPref(0), 
              server_info[1], server_info[2], server_info[3],
              server_info[4], password));
    }
  }
  list.sort(function (a,b) { return _strcmp(a.alias,b.alias); });
  return list;
}

function createServersMenuItem(server) {
  var cell = document.createElementNS(XUL_NS, 'menuitem');
  cell.setAttribute('id', server.id);
  cell.setAttribute('label', server.alias);
  cell.setAttribute('assocId', server.id);
  return cell;
}

function loadServerList()
{
    var menubox = document.getElementById("serversList");
    menubox.removeAttribute("collapsed");
    
    servers = getServerListFromPreference(globalPrefs);
    var menulist = document.getElementById("serverMenu");
    var menuitem;
    
    // add the default entry
    menuitem = document.createElementNS(XUL_NS, 'menuitem');
    if (FTPBrowserOnly) {
        menuitem.setAttribute('id', "blank_entry");
        menuitem.setAttribute('label', "");
    } else {
        menuitem.setAttribute('id', localDrives);
        menuitem.setAttribute('label', localDrives);
    }
    menulist.appendChild(menuitem);
    
    // add the deployment servers
    for (var i = 0; i < servers.length; i++) {
        menuitem = createServersMenuItem(servers[i]);
        menulist.appendChild(menuitem);
    }
    if (servers.length > 0) document.getElementById("serverMenuList").selectedIndex = 0
        
}

function clearHistoryList()
{
    var menulist = document.getElementById("lookInMenu");
    while (menulist.childNodes.length) {
        menulist.removeChild(menulist.firstChild);
    }
}

function clearServerList()
{
    var menulist = document.getElementById("serverMenu");
    while (menulist.childNodes.length) {
        menulist.removeChild(menulist.firstChild);
    }
}

function getAliasIndex(alias) {
  log.debug("getting the server object for "+alias);
  if (servers) {
  for (var i = 0; i < servers.length; i++) {
    if (alias == servers[i].alias) return i;
  }
  }
  return -1;
}

function onServerChanged(alias) {
    log.debug("onServerChanged("+alias+")");
    closeFTP();
    // just to allow closing the connection...
    if (!alias) {
        log.debug('onServerChanged set dir to null');
        setInitialDirectory(null);
        return;
    }
    currentServer = alias;
    var idx = getAliasIndex(alias);
    if (idx < 0) {
        if (alias == localDrives) {
            log.debug('onServerChanged clear the global ftp connection, set dir to '+lastLocalDirectory);
            setInitialDirectory(lastLocalDirectory);
            return;
        }
        if (alias.indexOf("ftp://") == 0) {
            log.debug('onServerChanged set dir to '+alias);
            setInitialDirectory(alias);
            return;
        }
        if (alias.indexOf("ftp.") == 0) {
            log.debug('onServerChanged set dir to '+alias);
            var url = "ftp://"+alias;
            setInitialDirectory(url);
            return;
        }
        dialog_alert("Unable to access URL\n");
        return;
    }
    log.debug('onServerChanged idx '+idx);
    var server = servers[idx];
    goServerConnect(server);
}

function goServerConnect(server) {
    // build url!
    var url = server.type.toLowerCase()+"://";
    if (server.username) {
        url += server.username;
        if (server.password) url += ":"+server.password;
        url += "@";
    }
    url += server.hostname;
    if (server.port) url += ":"+server.port;
    if (server.path[0] != "/") url += "/"+server.path;
    else url += server.path;
    log.debug('goServerConnect to '+url)
    setInitialDirectory(url);
}

function goServerPrefs() {
    // just load up prefs so they can add servers
    opener.prefs_doGlobalPrefs("serversItem");
    clearServerList();
    loadServerList();
}

function goHelp() {
    window.opener.launch_MainHelpTag(helpTag);
}
