/* Copyright (c) 2000-2001 ActiveState Corporation.
   See the file LICENSE.txt for licensing information. */

// Methods used by File settings and properties window
//
// Usage:
//  window.arguments[0].view    the view for which to present properties
//

//---- globals

const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

var log = getLoggingMgr().getLogger("currentViewPrefs");
//log.setLevel(LOG_DEBUG);
var widgets = null;  // hold references to commonly used DOM nodes
var gView = null;
var gExistingLineEndingsCache = null; // determining this is expensive

// A dictionary of XUL ids that have changed, i.e. the user changed some
// modifiable UI widget. This dictionary is used so OK/Apply knows what
// potential changes to look for.
var gChangedFields = {};

var gPrefWindow = null;

var gEncodingSvc = Components.classes["@activestate.com/koEncodingServices;1"].
                   getService(Components.interfaces.koIEncodingServices);



//---- routines for XUL

function OnLoad() {
    try {
        trace_onload();
        gView = window.arguments[0].view;
        if (typeof(gView) == "undefined" || !gView || !gView.prefs)
            throw "Could not locate a preference set in the window args!";

        // Set 'widgets' object, which contains references to interesting
        // elements in the dialog.
        widgets = new Object();
        widgets.dialog = document.getElementById("dialog");
        widgets.okButton = widgets.dialog.getButton("accept");
        widgets.cancelButton = widgets.dialog.getButton("cancel");
        widgets.tabbox = document.getElementById("fileTabbox");
        widgets.tabs = {
            "properties": document.getElementById("properties_tab"),
            "scc": document.getElementById("scc_tab"),
            "editing": document.getElementById("editing_tab"),
            "indentation": document.getElementById("indentation_tab"),
            "preview": document.getElementById("preview_tab")
        }
        widgets.encoding = document.getElementById("encoding");
        widgets.lineEndings = document.getElementById("lineEndings");
        widgets.bom = document.getElementById("bom");
        widgets.preserveLineEndings = document.getElementById("preserveLineEndings");
        widgets.mixedLineEndings = document.getElementById("mixedLineEndings");
        widgets.readonly = document.getElementById("readonly");
        widgets.language = document.getElementById("language");
        widgets.preview = document.getElementById("preview");
        widgets.changePreview = document.getElementById("change-preview");
        widgets.clearPreview = document.getElementById("clear-preview");

        // Pretty up the buttons.
        widgets.okButton.setAttribute("accesskey", "O");
        widgets.cancelButton.setAttribute("accesskey", "C");

        // Setup the title.
        window.title = "'"+gView.title+"' Properties and Settings";

        // Prep the magic-pref-handling stuff for settings that are stored
        // if this view's prefs.
        gPrefWindow = new koPrefWindow(null, gView.prefs);
        if(!gPrefWindow) throw "failed to create prefwindow";
        gPrefWindow.onsinglepageload();

        initFileProperties();
        initDocumentSettings();



        initPreview();
        initViewDependentUI();

        // Switch to the properties tab (the first one).
        widgets.tabbox.selectedItem = widgets.tabs["properties"];

        //handleLanguageSensitivities();
    } catch(ex) {
        log.exception(ex, "Error loading current view prefs dialog.");
    }
}


// This function handles the disabled/enabled state of UI elements 
// depending on the language (or other view attributes).
function initViewDependentUI()  {
    //XXX This handling is *kind of* bogus, but it works
    var linterLanguageService = gView.document.languageObj.
          getLanguageService(Components.interfaces.koILinterLanguageService);

    if (!linterLanguageService) {
        document.getElementById('editUseLinting')
                .setAttribute('disabled', 'true');
    }
    var scimoz = gView.scimoz;
    document.getElementById('useTabs').setAttribute('checked',
                        scimoz.useTabs ? 'true' : 'false');
    document.getElementById('indentWidth').value = String(scimoz.indent);
    document.getElementById('tabWidth').value = String(scimoz.tabWidth);
    document.getElementById('showIndentationGuides').setAttribute('checked',
                        scimoz.indentationGuides ? 'true' : 'false');
    document.getElementById('showEOL').setAttribute('checked',
                        scimoz.viewEOL ? 'true' : 'false');
    document.getElementById('showWhitespace').setAttribute('checked',
                        scimoz.viewWS ? 'true' : 'false');
    document.getElementById('showLineNumbers').setAttribute('checked',
                        scimoz.getMarginWidthN(0) != 0 ? 'true' : 'false');
    document.getElementById('editUseAutoWrap').setAttribute('checked',
                        scimoz.wrapMode == scimoz.SC_WRAP_WORD ? 'true' : 'false');
    
    // Code Intelligence-related stuff
    var langObj = gView.document.languageObj;
    var completionLangSvc;



    completionLangSvc = langObj.getLanguageService(Components.interfaces.koICompletionLanguageService)
            || langObj.getLanguageService(Components.interfaces.koICodeIntelCompletionLanguageService);

    if (!completionLangSvc)  {
        document.getElementById("codeintel_completion_triggering_enabled")
                .setAttribute("disabled", "true");
    }
    if (!opener.gCodeIntelSvc
        || !opener.gCodeIntelSvc.isSupportedLanguage(langObj.name))
    {
        document.getElementById("codeintel_rescan_while_typing_enabled")
                .setAttribute("disabled", "true");
    }
}

// Should be called whenever a UI element is modified by the user.
// This data is used to actually write out the changes on Apply/OK.
function setField(name, value)
{
    gChangedFields[name] = value;
}


function initFileProperties()
{
    try {
        var name, dirname, size, ctime, mtime, atime, readonly;

        if (!gView.document.file) {
            name = gView.title;
            dirname = size = ctime = mtime = atime = "N/A";
            readonly = null; // means disabled (N/A)
        } else {
            var file = gView.document.file;
            name = file.baseName;

            dirname = file.dirName;
            size = file.fileSize + " bytes";

            // Time info.
            var timeSvc = Components.classes["@activestate.com/koTime;1"].
                            getService(Components.interfaces.koITime);
            var prefSvc = Components.classes['@activestate.com/koPrefService;1'].
                            getService(Components.interfaces.koIPrefService);
            var format = prefSvc.prefs.getStringPref("defaultDateFormat");
            var timeTuple;
            timeTuple = timeSvc.localtime(file.createdTime, new Object());
            ctime = timeSvc.strftime(format, timeTuple.length, timeTuple);
            timeTuple = timeSvc.localtime(file.lastModifiedTime, new Object());
            mtime = timeSvc.strftime(format, timeTuple.length, timeTuple);
            timeTuple = timeSvc.localtime(file.lastAccessedTime, new Object());
            atime = timeSvc.strftime(format, timeTuple.length, timeTuple);

            // Determine read-only status.
            readonly = file.isReadOnly;
        }
    
        // Update UI with data.
        document.getElementById("properties-name").value = name;



        document.getElementById("location").value = dirname;
        document.getElementById("size").value = size;



        document.getElementById("modified").value = mtime;
        document.getElementById("accessed").value = atime;
        if (readonly == null) { // i.e. N/A
            widgets.readonly.checked = false;
            widgets.readonly.setAttribute("tooltiptext", "Not Applicable");
            widgets.readonly.setAttribute("disabled", "true");
        } else {
            widgets.readonly.checked = readonly;
            if (widgets.readonly.hasAttribute("disabled"))
                widgets.readonly.removeAttribute("disabled");
            if (widgets.readonly.hasAttribute("tooltiptext"))
                widgets.readonly.removeAttribute("tooltiptext");
        }
    } catch(ex) {
        log.exception(ex, "Error updating file properties.");
    }
}


function initDocumentSettings()
{
    try {
        // Language
        widgets.language.selection = gView.document.language;

        // Encoding
        initEncoding();

        // Line endings
        var nle = gView.document.new_line_endings;
        if (nle == gView.document.EOL_LF) {
            widgets.lineEndings.value = "EOL_LF";
        } else if (nle == gView.document.EOL_CR) {
            widgets.lineEndings.value = "EOL_CR";
        } else if (nle == gView.document.EOL_CRLF) {
            widgets.lineEndings.value = "EOL_CRLF";
        }

        gExistingLineEndingsCache = gView.document.existing_line_endings;
        if (gExistingLineEndingsCache == gView.document.EOL_MIXED) {
            if (widgets.mixedLineEndings.hasAttribute("collapsed"))
                widgets.mixedLineEndings.removeAttribute("collapsed");
        } else {
            widgets.mixedLineEndings.setAttribute("collapsed", "true");
        }
        // Always preserve the current document by default.
        widgets.preserveLineEndings.checked = true;

    } catch(ex) {
        log.exception(ex, "Error updating settings.");
    }
}


function resetLanguage() {
    try {
        var registry = Components.classes["@activestate.com/koLanguageRegistryService;1"].
                       getService(Components.interfaces.koILanguageRegistryService);
        var language = registry.suggestLanguageForFile(gView.title);
        widgets.language.selection = language;
        setField("language", "");  // empty string means to reset
    } catch(ex) {
        log.exception(ex, "Error resetting language.");
    }
}


function initEncoding()  {
    var enc = gView.document.encoding;  // koIEncoding object
    var encIndex = gEncodingSvc.get_encoding_index(enc.python_encoding_name);

    // Build the menupopup.
    var menupopup = encodingMenu_buildHierarchy(gEncodingSvc.encoding_hierarchy,
                                                true,
                                                'changeEncoding(this)'); // action
    if (encIndex == -1) {
        // Current encoding is not in the std encoding list -- prepend it.
        var menuitem = document.createElementNS(XUL_NS, 'menuitem');
        menuitem.setAttribute('data', enc.python_encoding_name);
        menuitem.setAttribute('label', enc.friendly_encoding_name);
        menuitem.setAttribute('oncommand', 'changeEncoding(this)');
        menupopup.insertBefore(menuitem, menupopup.childNodes[0]);
    }
    widgets.encoding.appendChild(menupopup);

    // Make sure current is selected.
    if (encIndex == -1) {
        widgets.encoding.selectedIndex = 0;
    } else {
        var item = document.getElementsByAttribute('data',
                    enc.python_encoding_name)[0];
        widgets.encoding.selectedItem = item;
    }

    // Init BOM checkbox.
    widgets.bom.checked = enc.use_byte_order_marker;
    updateBOM(enc.encoding_info);
}


function updateBOM(encodingInfo)
{
    if (encodingInfo.byte_order_marker != '') {
        if (widgets.bom.hasAttribute("disabled"))
            widgets.bom.removeAttribute("disabled");
    } else {
        widgets.bom.setAttribute('disabled', 'true');
    }
}


function changeEncoding(item)
{
    widgets.encoding.selectedItem = item;
    var encodingName = item.getAttribute("data");
    updateBOM(gEncodingSvc.get_encoding_info(encodingName));
    setField("encoding_and_or_bom", encodingName);
}


function initPreview() {
    // Browser preview requires a saved document.
    if (gView.document.isUntitled) {
        widgets.preview.setAttribute("disabled", "true");
        widgets.changePreview.setAttribute("disabled", "true");
        widgets.clearPreview.setAttribute("disabled", "true");
    }
    
    else {
        if (gView.prefs.hasStringPref("preview_path")) {
            widgets.preview.value = gView.prefs.getStringPref("preview_path");
        }
        updateClearPreviewButton();
    }
}

function changePreview()
{
    var pick = dialog_pickPreview(gView.document.file.displayPath,
                                  null, // language
                                  "setting"); // mode
    if (pick == null)
        return;
    widgets.preview.value = pick.preview;
    setField("preview", pick.preview);
    updateClearPreviewButton();
    widgets.okButton.focus();
}

function clearPreview()
{
    widgets.preview.value = null;
    setField("preview", null);
    updateClearPreviewButton();
    widgets.okButton.focus();
}

function updateClearPreviewButton()
{
    if (!widgets.preview.value) {
        widgets.clearPreview.setAttribute("disabled", "true");
    } else {
        if (widgets.clearPreview.hasAttribute("disabled"))
            widgets.clearPreview.removeAttribute("disabled");
    }
}

























// Apply the given encoding and BOM changes to the current document.
//
// Returns true iff any changes were successfully made.
//
function applyEncodingAndBOM(encodingName, bom)
{
    // Short-circuit if the encoding name is the same.
    if (encodingName == gView.document.encoding.python_encoding_name) {
        gView.document.encoding.use_byte_order_marker = bom;
        return true;
    }

    var enc = Components.classes["@activestate.com/koEncoding;1"].
                     createInstance(Components.interfaces.koIEncoding);
    enc.python_encoding_name = encodingName;
    enc.use_byte_order_marker = bom;

    var warning = gView.document.languageObj.getEncodingWarning(enc);
    var question = warning+" Are you sure that you want to change the encoding?";
    if (warning == "" || dialog_yesNo(question, "No") == "Yes") {
        try {
            gView.document.encoding = enc;
            // and reset the linting
            gView.lintBuffer.request();
        } catch(ex) {
            var err;
            var lastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"].
                               getService(Components.interfaces.koILastErrorService);
            var errno = lastErrorSvc.getLastErrorCode();
            var errmsg = lastErrorSvc.getLastErrorMessage();
            if (errno == 0) {
                // koDocument.set_encoding() says this is an internal error
                err = "Internal error setting the encoding for '"+
                          gView.document.displayPath+"' to '"+
                          encodingName+"'."
                dialog_internalError(err, err+"\n\n"+errmsg+
                                          "\n\n\nTRACEBACK:\n"+
                                          gTimeline.getStack(),
                                     opener);
            } else {
                question = errmsg+"  Komodo can force the conversion by "+
                           "replacing all unrepresentable "+
                           "characters with a question mark, '?', or you "+
                           "may Cancel and select a different encoding.";
                var choice = dialog_customButtons(question,
                                                  ["Force", "Cancel"],
                                                  "Cancel"); // default
                if (choice == "Force") {
                    try {
                        gView.document.forceEncodingFromEncodingName(encodingName);
                    } catch (ex2) {
                        err = "Internal error forcing the encoding for '"+
                                  gView.document.baseName+"' to '"+
                                  encodingName+"'."
                        dialog_internalError(err, err+"\n\n"+errmsg+
                                                  "\n\n\nTRACEBACK:\n"+
                                                  gTimeline.getStack(),
                                             opener);
                        return false;
                    }
                } else {
                    return false;
                }
            }
        }
        return true;
    }
    
    return false;
}


function OK() {
    try {
        var lastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"].
                           getService(Components.interfaces.koILastErrorService);
        var observerSvc = Components.classes["@mozilla.org/observer-service;1"].
                          getService(Components.interfaces.nsIObserverService);

        // Update changed fields.
        if (!gPrefWindow.onSinglePageOK()) {
            dialog_alert("Error saving 'View', 'Editing' or "+
                         "'Indentation' settings");
            return false;
        }

        var field, value;
        for (field in gChangedFields) {
            value = gChangedFields[field];
            log.info("updating changed field '"+field+"' to '"+value+"'\n");
            switch(field) {
            case "readonly":
                if (value != gView.document.file.isReadOnly) {
                    var osSvc = Components.classes["@activestate.com/koOs;1"].
                                 getService(Components.interfaces.koIOs);
                    var path = gView.document.file.displayPath;
                    try {
                        osSvc.setWriteability(path, !value);
                    } catch(ex) {
                        dialog_alert("Error setting read-only mode for '"
                                     +gView.title+"': "+
                                     lastErrorSvc.getLastErrorMessage());
                        widgets.readonly.focus();
                        return false;
                    }
                    observerSvc.notifyObservers(null, "file_status_now",
                                                gView.document.file.URI);
                }
                break;

            case "language":
                if (value == "" || value != gView.document.language) {
                    gView.document.language = value;
                    observerSvc.notifyObservers(gView,
                                                "current_view_language_changed",
                                                null);
                }
                break;

            case "lineEndings":
                var le;
                if (value == "EOL_LF") {
                    le = gView.document.EOL_LF;
                } else if (value == "EOL_CR") {
                    le = gView.document.EOL_CR;
                } else if (value == "EOL_CRLF") {
                    le = gView.document.EOL_CRLF;
                }
                // If we've changed things, save it as a pref.
                if (gView.document.new_line_endings != le) {
                    var eolpref = '';
                    switch (le) {
                        case gView.document.EOL_LF: eolpref = 'LF'; break;
                        case gView.document.EOL_CR: eolpref = 'CR'; break;
                        case gView.document.EOL_CRLF: eolpref = 'CRLF'; break;
                        default:
                            log.error("unknown line ending: " + le);
                    }
                    if (eolpref) {
                        gView.prefs.setStringPref('endOfLine', eolpref);
                    }
                }
                // Make the change even if 'le' is unchanged to allow a
                // change to "Preserve Line Endings" on EOL_MIXED files to
                // correct the file.
                gView.document.new_line_endings = le;
                if (! widgets.preserveLineEndings.checked) {
                    // This changes the current document's line endings.
                    gView.document.existing_line_endings = le;
                }
                break;

            case "encoding_and_or_bom":
                var encodingName = widgets.encoding.selectedItem.getAttribute("data");
                var bom = ((!widgets.bom.getAttribute("disabled") || widgets.bom.getAttribute("disabled") == "false")
                           && widgets.bom.checked);
                var ok = applyEncodingAndBOM(encodingName, bom);
                if (!ok) return false;
                observerSvc.notifyObservers(gView,
                                            "current_view_encoding_changed",
                                            null);
                break;

            case "preview":
                if (value == null) {
                    if (gView.prefs.hasStringPref("preview_path"))
                        gView.prefs.deletePref("preview_path")
                } else {
                    gView.prefs.setStringPref("preview_path", value);
                }
                break;

            default:
                log.error("Don't know how to update changed field '"+field+"'.");
            }
        }
        
        window.arguments[0].retval = "OK";
        return true;
    } catch (ex) {
        log.exception(ex, "Error OK'ing dialog.");
    }
    return false;
}


function Cancel() { 
    window.arguments[0].retval = "Cancel";
    return true;
}


