/* Copyright (c) 2000-2003 ActiveState Corporation.
   See the file LICENSE.txt for licensing information. */

// start the extension
registeredExtensions[registeredExtensions.length] = new peFile();

function peFile() {
    this.name = 'peFile';
    this.koLanguageRegistryService =
                Components.classes["@activestate.com/koLanguageRegistryService;1"].
                            getService(Components.interfaces.koILanguageRegistryService);
    this.observerSvc = Components.classes["@mozilla.org/observer-service;1"].
                       getService(Components.interfaces.nsIObserverService);
    this.sysUtilsSvc = Components.classes["@activestate.com/koSysUtils;1"].
                    getService(Components.interfaces.koISysUtils);
    this.lastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"].
                        getService(Components.interfaces.koILastErrorService);
}

// The following two lines ensure proper inheritance (see Flanagan, p. 144).
peFile.prototype.constructor = peFile;

peFile.prototype.init = function() {
    // register our command handlers
    this.observerSvc.addObserver(this, 'file_status',false);
    this.observerSvc.addObserver(this, 'file_status_updated',false);

    extensionManager.setDatapoint('Date','lastModifiedTime');
    extensionManager.setDatapoint('Size','fileSize');
    extensionManager.setDatapoint('Status','sccStatus');
    extensionManager.setDatapoint('Rev','sccLocalRevision');
    extensionManager.setDatapoint('Depot Rev','sccDepotRevision');
    extensionManager.setDatapoint('Action','sccAction');
    this._clipboard = [];
}

peFile.prototype.registerCommands = function() {
    extensionManager.registerCommand('cmd_openInDir', this);
    extensionManager.registerCommand('cmd_dumpPartProperties', this);
    extensionManager.registerCommand('cmd_openFilePart', this);
    extensionManager.registerCommand('cmd_makeDirectoryShortcutFromFile', this);
    extensionManager.registerCommand('cmd_refreshStatus', this);
    extensionManager.registerCommand('cmd_editProperties', this);
    extensionManager.registerCommand('cmd_showUnsavedChanges', this);
    extensionManager.registerCommand('cmd_compareFiles', this);
    extensionManager.registerCommand('cmd_compareFileWith', this);
    extensionManager.registerCommand("cmd_exportItems",this);
    extensionManager.registerCommand("cmd_exportToPackage",this);
    extensionManager.registerCommand("cmd_cut",this);
    extensionManager.registerCommand("cmd_copy",this);
    extensionManager.registerCommand("cmd_paste",this);
    extensionManager.registerCommand('cmd_rename', this);
}

peFile.prototype.registerEventHandlers = function() {
    extensionManager.addEventHandler(Components.interfaces.koIPart_file,'ondblclick',this);
    extensionManager.addEventHandler(Components.interfaces.koIDirectoryShortcut,'ondblclick',this);
    // we don't add a dblclick handler for Components.interfaces.koIPart_ProjectRef, since
    // it inherits from koIPart_file
    
    extensionManager.addEventHandler(Components.interfaces.koIPart_file,'onUpdatePart',this);
    extensionManager.addEventHandler(Components.interfaces.koIProject,'onUpdatePart',this);
    extensionManager.addEventHandler(Components.interfaces.koIPart_ProjectRef,'onUpdatePart',this);
    extensionManager.addEventHandler(Components.interfaces.koIDirectoryShortcut,'onUpdatePart',this);
}

peFile.prototype.registerMenus = function() {
    extensionManager.createMenuItem(Components.interfaces.koIPart_file,
                                    'Open File','cmd_openFilePart',
                                    null,
                                    null,
                                    true);
    extensionManager.createMenuItem(Components.interfaces.koIDirectoryShortcut,
                                    'Open...',
                                    'cmd_openInDir',
                                    null,
                                    null,
                                    true /* primary */);
    extensionManager.createMenuItem(Components.interfaces.koIPart_file,
                                    'Make "Open..." Shortcut','cmd_makeDirectoryShortcutFromFile');
    extensionManager.createMenuItem(Components.interfaces.koIPart_file,
                                    'Refresh Status','cmd_refreshStatus');
    extensionManager.createMenuItem(Components.interfaces.koIPart_file,
                                    'Show unsaved changes','cmd_showUnsavedChanges');
    extensionManager.createMenuItem(Components.interfaces.koIPart_file,
                                    'Compare Files...','cmd_compareFiles');
    extensionManager.createMenuItem(Components.interfaces.koIPart_file,
                                    'Compare File With...','cmd_compareFileWith');
    extensionManager.createMenuItem(Components.interfaces.koIPart,
                                    'Export as Project File...','cmd_exportItems');
    extensionManager.createMenuItem(Components.interfaces.koIPart,
                                    'Export Package...','cmd_exportToPackage');

    extensionManager.createMenuItem(Components.interfaces.koIPart,
                                    'Rename','cmd_rename');
    // XXX debug info
    extensionManager.createMenuItem(Components.interfaces.koIPart,'Dump',
                                    'cmd_dumpPartProperties');
}

peFile.prototype.supportsCommand = function(command, item) {
    var items = null;
    if (gFocusedProjectView) {
        items = gFocusedProjectView.getSelectedItems();
    }
    switch (command) {
    case 'cmd_cut':
    case 'cmd_copy':
        if (!items || items.length == 0) return false;
        if (command == 'cmd_cut' && !gFocusedProjectView.manager.writeable()) return false;
        // We want to make sure that no projects are selected
        for (var i = 0; i < items.length; i++) {
            if (items[i].type == 'project') {
                return false;
            }
        }
        return true;
    case 'cmd_paste':
        return 1; // see http://bugs.activestate.com/Komodo/show_bug.cgi?id=27527

    case 'cmd_rename':
        // we only allow renaming parts that are not attached to a url,
        // such as a virtual folder
        //dump('asked about rename\n');
        var havemultiple = items && items.length > 1;
        if (! gFocusedProjectView.manager.writeable()) return false;
        if (!item && gFocusedProjectView)
            item = gFocusedProjectView.getSelectedItem();
        //dump('updating cmd_rename, returning ' + (!havemultiple && item && item.renameable) + '\n');
        return !havemultiple && item && item.renameable;
    case 'cmd_dumpPartProperties':
        // Only return true in developer builds
        var infoSvc = Components.classes["@activestate.com/koInfoService;1"].
                  getService(Components.interfaces.koIInfoService);
        return infoSvc.installerType == '';
    case 'cmd_openFilePart':
        for (i = 0; i < items.length; i++) {
            if (items[i].type == 'file') return true;
        }
        return false;
    case 'cmd_openInDir':
        return (items.length == 1 && items[0].type == 'DirectoryShortcut');
    case 'cmd_makeDirectoryShortcutFromFile':
        if (! gFocusedProjectView.manager.writeable()) return false;
        return (items.length == 1 && items[0].type == 'file' && items[0].url.indexOf('file://') == 0);
    case 'cmd_refreshStatus':
        // if a toolbox has focus, get the currently selected item and refresh it
        // otherwise, always refresh the currentView if there is one.
        if (view_getFocusedProjectView()) {
            item = gFocusedProjectView.getSelectedItem();
            // item must be a file
            if (item && item.url) return true;
        }
        return gViewMgr.currentView != null;
        break;
    case 'cmd_editProperties':
        if (!items || items.length < 1) return false;
        for (i = 0; i < items.length; i++) {
            type = items[i].part.type;
            switch (type) {
                case 'file':
                case 'folder':
                case 'project':
                case 'DirectoryShortcut':
                case 'URL':
                case 'command':
                case 'macro':
                case 'snippet':
                case 'template':
                case 'webservice':
                    // leave in so we have support for old project files
                case 'menu':
                case 'toolbar':
                    continue;
                default:
                    //dump('returning false because type is not correct\n');
                    return false;
            }
        }
        return true;
    case 'cmd_showUnsavedChanges':
        view = gViewMgr.currentView;
        return (view && view.document && view.document.isDirty &&
                !view.document.isUntitled);
    case 'cmd_compareFiles':
        return (items.length == 2 && items[0].type == 'file' && items[1].type == 'file');
    case 'cmd_compareFileWith':
        return (items && items.length == 1 && items[0].type == 'file');
    case 'cmd_exportItems':
        // the following isn't great, but it deals with
        // http://bugs.activestate.com/Komodo/show_bug.cgi?id=23226
        // with minimal risk.
        if (items && items.length == 1 && items[0].type == 'project') return 0;
        return items.length > 0;
    case 'cmd_exportToPackage':
        return items && items.length > 0;
    default:
        break;
    }
    return false;
}

peFile.prototype.isCommandEnabled = peFile.prototype.supportsCommand;

peFile.prototype.doCommand = function(command) {
    var item, items = null;
    var i, fname, otherfile;
    var dirname = '';
    if (gFocusedProjectView) {
        item = gFocusedProjectView.getSelectedItem();
        items = gFocusedProjectView.getSelectedItems();
    }
    switch (command) {
    case 'cmd_rename':
        if (!item) return;
        gFocusedProjectView.manager.renameItem(item);
        break;
    case 'cmd_openFilePart':
        // get the current selection, then open the file
        if (!items) return;
        var paths = [];
        for (i = 0; i < items.length; i++) {
            if (items[i].type == 'file') paths.push(items[i].url);
        }
        open_openMultipleURIs(paths);
        break;
    case 'cmd_makeDirectoryShortcutFromFile':
        var view, url;
        dirname = null;
        if (typeof(item) != 'undefined' && item) {
            url = item.part.getStringAttribute('url');
            var pathSvc = Components.classes["@activestate.com/koOsPath;1"]
                                        .getService(Components.interfaces.koIOsPath);
            dirname = uriparse_URIToLocalPath(url);
            if (item.part.type != 'folder') {
                dirname = pathSvc.dirname(dirname);
            }
        }
        peFile_addDirectoryShortcut(item, dirname);
        break;
    case 'cmd_refreshStatus':
        doRefreshStatus();
        break;
    case 'cmd_dumpPartProperties':
        if (!gFocusedProjectView) return;
        item = gFocusedProjectView.getSelectedItem();
        if (item) item.dump();
        break;
    case 'cmd_openInDir':
        if (!gFocusedProjectView) return;
        item = gFocusedProjectView.getSelectedItem();
        if (item) {
            OpenDirectoryShortcut(item.part)
        }
        break;
    case 'cmd_editProperties':
        if (!view_getFocusedProjectView())
            return;
        if (!items) return;
        for (i = 0; i < items.length; i++) {
            item = items[i];
            part = item.part;
            switch (part.type) {
                case 'webservice':
                    // leave in so we have support for old project files
                case 'URL':
                    peURL_editProperties(item, i);
                    continue;
                case 'command':
                    command_editProperties(item, i);
                    continue;
                case 'macro':
                    macro_editProperties(item, i);
                    continue;
                case 'snippet':
                    snippet_editProperties(item, i);
                    continue;
                case 'DirectoryShortcut':
                    this.editDirectoryShortcut(item, i);
                    continue;
                case 'template':
                    template_editProperties(item, i);
                    continue;
                case 'file':
                case 'folder':
                case 'project':
                    this._fileProperties(item, i);
                    continue;
                case 'menu':
                case 'toolbar':
                    peMenu_editProperties(item, i);
                    continue;
            }
        }
        break;
    case 'cmd_showUnsavedChanges':
        view = gViewMgr.currentView;
        var changes = view.document.getUnsavedChanges();
        launch_DiffWindow(changes, 'Unsaved changes for ' + view.document.displayPath);
        break;
    case 'cmd_compareFiles':
        if (!gFocusedProjectView || !items || items.length == 0) {
            return;
        } else {
            fname = items[0].file.URI;
            otherfile = items[1].file.URI;
        }
        peFile_ShowDiffs(fname, otherfile);
        return;
    case 'cmd_compareFileWith':
        if (!gFocusedProjectView) return;
        fname = items[0].file.URI;
        var pickerDir = items[0].file.isLocal? items[0].file.dirName : '';
        otherfile = filepicker_openFile(pickerDir);
        if (otherfile) {
            peFile_ShowDiffs(fname, otherfile);
        }
        return;
    case "cmd_exportItems":
        if (!gFocusedProjectView) return;
        Toolbox_ExportItems(items);
        break;
    case 'cmd_exportToPackage':
        Toolbox_ExportPackageItems(items);
        break;
    case 'cmd_cut':
        // get the current selection, then open the file
        if (!gFocusedProjectView) return;
        this.doCut(items)
        break;
    case 'cmd_copy':
        // get the current selection, then open the file
        this.doCopy(items)
        break;
    case 'cmd_paste':
        if (this._clipboard.length == 0) return;
        this.doPaste(items)
        break;
    default:
        break;
    }
}

function peFile_addDirectoryShortcut(item, dirname)
{
    var parentindex, index;
    var dirshortcut = Components.classes["@activestate.com/koPart?type=DirectoryShortcut;1"]
        .createInstance(Components.interfaces.koIDirectoryShortcut);
    dirshortcut.value = dirname;
    if (dirname.length > 25) {
        dirname = dirname.substr(0, 10) + '...' + dirname.substr(dirname.length - 15, dirname.length)
    }
    dirshortcut.setStringAttribute('name', dirname);
    if (item.isContainer) {
        parentindex = gFocusedProjectView.getIndexByPart(item);
    } else {
        index = gFocusedProjectView.getIndexByPart(item);
        parentindex = gFocusedProjectView.getParentIndex(index);
    }
    gFocusedProjectView.manager.addItem(dirshortcut, parentindex);
}

function _openDiffWindowForFiles(fname1, fname2) {
    window.setCursor("wait");
    try {
        var diff = null;
        try {
            docSvc = Components.classes['@activestate.com/koDocumentService;1']
                    .getService(Components.interfaces.koIDocumentService);
            diff = docSvc.diffFiles(fname1, fname2);
        } catch (ex) {
            log.error(ex);
            var msg = "Error calculating differences: " +
                      this.lastErrorSvc.getLastErrorMessage();
            alert(msg);
            return;
        }
        if (!diff) {
            dialog_alert("There are no differences between '"+fname1+
                         "' and '"+fname2+"'.");
        } else {
            launch_DiffWindow(diff, "Comparison of '"+fname1+"' and '"+fname2+"'");
        }
    } finally {
        window.setCursor('auto');
    }
}


peFile.prototype.ondblclick = function(item,event) {
    if (item.type == 'DirectoryShortcut') {
        OpenDirectoryShortcut(item.part);
    } else {
        open_openURI(item.url);
    }
}

peFile.prototype.onUpdatePart = function(item,event)
{
    if (item.type == 'DirectoryShortcut') {
        item.renameable = true;
        item.iconstyle['DirectoryShortcut'] = 1;
        return;
    }

    var useprojecticon = false;
    try {
        useprojecticon = item.part.QueryInterface(Components.interfaces.koIProject) != null;
        item.renameable = false;
    } catch (e) {}    
    if (!useprojecticon) {
        try {
            useprojecticon = item.part.QueryInterface(Components.interfaces.koIPart_ProjectRef) != null;
        } catch (e) {}    
    }
    // event is null for onUpdatePart
    // get the file properties!
    if (useprojecticon) {
        item.iconstyle['project'] = 1;
        item.iconopenstyle['project'] = 1;
    } else {
        item.iconstyle['projectFile'] = 1;
    }
    if (typeof(item.properties['filesize']) == 'undefined') {
        this.getDefaultFileProperties(item);
    }
}

peFile.prototype.getDefaultFileProperties = function(item)
{
    item.properties['size'] = 0;
    item.properties['read'] = true;
    item.properties['write'] = true;
    item.properties['ctime']  = "Not Available";
    item.properties['mtime']  = "Not Available";
    item.properties['atime']  = "Not Available";



    item.iconopenstyle['readwrite']=item.iconstyle['readwrite']=0;
    item.iconopenstyle['readonly']=item.iconstyle['readonly']=0;
}

// XXX duplicated from currentViewPrefs.js
function getDateTimeString(datetime) {
    // XXX not localization friendly
    var now = new Date();
    var months = ["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];
    var mins = datetime.getMinutes();
    if (mins.toString().length == 1)  {
        mins = '0' + mins;
    }
    var hours = datetime.getHours();
    var fulltime;
    if (hours > 12)  {
        hours = hours-12;
        fulltime = hours.toString() + ':' + mins + " PM";
    }
    else  {
        fulltime = hours.toString() + ':' + mins + " AM";
    }
    var dtstring;
    if (datetime.getFullYear() == now.getFullYear() &&
        datetime.getMonth() == datetime.getMonth() &&
        now.getDate() == datetime.getDate()) {
            dtstring = fulltime;
    } else if (datetime.getFullYear() == now.getFullYear()) {
        dtstring = (datetime.getMonth()+1)+'/'+datetime.getDate()+'/'+datetime.getFullYear();
    } else {
        dtstring = (datetime.getMonth()+1)+'/'+datetime.getFullYear();
    }
    return dtstring;
}

function updateStatusForURL(url) {
    var item;
    var i;
    // get the item from the url
    var items = findItemsByURL(url);
    //dump(items.length+' matching parts '+url+'!\n');
    for (i=0; i < items.length; i++) {
        items[i].updateStatus();
    }
}

function doRefreshStatus () {
    var urls = {};
    var url;
    var item, items;
    
    if (view_getFocusedProjectView()) {
        // gProjectManager.log.debug("project or toolbox has focus\n");
        items = gFocusedProjectView.getSelectedItems();
        // otherwise, look at the focused project view's selected parts.
        if (items.length == 1 && items[0].type != 'project') {
            urls[items[0].url]=1;
        } else {
            for (i=0; i<items.length; i++) {
                items[i].updateUniquePaths(urls,true);
            }
        }
    } else {
        // always fall back to the current view
        var view = gViewMgr.currentView;
        item = view.item;
        if (!view.document || view.document.isUntitled) {
            // This is an unsaved file -- cannot refresh.
            return;
        }
        item.url = view.document.file.URI;
        urls[item.url]=1;
        
        // Also refresh CodeIntel data for this view, if enabled.
        if (view.codeIntelEnabled) {
            // fake an edit to get a re-scan of the document
            gCodeIntelSvc.ideEvent_EditedCurrentDocument(
                view.document,
                view.scimoz,
                0, // no lines added
                true); // yes, rescan
        }
    }
    try {
        for (url in urls) {
            // gProjectManager.log.debug("updating status for url: "+url+"\n");
            gObserverSvc.notifyObservers(this,'file_status_now',url);
        }
    } catch(e) { /* noone listening, ignore error */
        // gProjectManager.log.error(e);
    }
}

peFile.prototype._fileProperties = function(item, offset)
{
    if (typeof(item.properties['missing']) != 'undefined' && item.properties['missing']) {
        alert("The file " + item.name + " does not exist on disk at " +
              item.file.displayPath);
    } else {
        var obj = new Object();
        obj.item = item;
        obj.task = 'edit';
        obj.offsetIndex = offset;
        openWindowMultipleInstance(
            "chrome://jaguar/content/project/fileProperties.xul", 
            "komodo_fileproperties",
            "chrome,close=yes,dependent=yes,resizable=yes", obj);
    }
}

peFile.prototype.editDirectoryShortcut = function(item) {
    var obj = new Object();
    obj.item = item;
    obj.task = 'edit';
    obj.imgsrc = 'chrome://jaguar/skin/images/open.png';
    obj.type = 'DirectoryShortcut';
    obj.prettytype = 'Directory Shortcut';
    openWindowMultipleInstance(
        "chrome://jaguar/content/project/simplePartProperties.xul", 
        "komodo_DirectoryShortcutProperties",
        "chrome,close=yes,dependent=yes,resizable=yes", obj);
}

peFile.prototype.observe = function (object, topic, urllist)
{
    //dump('peFile got '+topic+' for '+url+'\n');
    switch (topic) {
        case "file_status":
            urllist = urllist.split('\n');
            for (var i=0; i < urllist.length; ++i) {
                updateStatusForURL(urllist[i]);
            }
            break;
        case "file_status_updated":
            for (var i=0; i < gPartManagers.length; i++) {
                gPartManagers[i].refreshView();
            }
            break;
        default:
            // ignore
            break;
    }
}

peFile.prototype.doCut = function(items)
{
    var needUpdate = this._clipboard.length == 0 && items.length != 0;
    this._clipboard = [];
    var i, index, parent;
    // Find out what the "first" index of the parts is
    var first_index = null;
    for (i = 0; i < items.length; i++) {
        index = gFocusedProjectView.getIndexByPart(items[i]);
        if (first_index == null || index < first_index) {
            first_index = index;
        }
    }
    var parentindex = gFocusedProjectView.getParentIndex(first_index);
    // Do the cut
    for (i = 0; i < items.length; i++) {
        this._clipboard.push(items[i].clone());
        gFocusedProjectView.manager.removeItem(items[i], true /* skip dialog */);
    }
    // Select the parent
    if (parentindex != -1) {
        gFocusedProjectView.selectIndex(parentindex);
    }
    if (needUpdate) {
        window.updateCommands('clipboard');
    }
}

peFile.prototype.doCopy = function(items)
{
    var needUpdate = this._clipboard.length == 0 && items.length != 0;
    this._clipboard = [];
    for (var i = 0; i < items.length; i++) {
        this._clipboard.push(items[i].clone());
    }
    if (needUpdate) {
        window.updateCommands('clipboard');
    }
}

peFile.prototype.doPaste = function(items)
{
    // Find the parent container
    var parent, index, parentindex;
    var item = gFocusedProjectView.getSelectedItem();
    // If there is no selected item, then see if we're in the toolbox -- at which point
    // the toolbox itself is the 'item'
    if (!item) {
        if (gFocusedProjectView == toolboxMgr.viewMgr.view) {
            parent = toolboxMgr.wrappedproject;
            parentindex = gFocusedProjectView.getIndexByPart(parent);
        } else {
            return;
        }
    } else {
        if (item.isContainer) {
            parentindex = gFocusedProjectView.getIndexByPart(item);
            parent = item;
        } else {
            index = gFocusedProjectView.getIndexByPart(item);
            parentindex = gFocusedProjectView.getParentIndex(index);
        }
    }
    for (var i = 0; i < this._clipboard.length; i++) {
        item = this._clipboard[i];
        if (item.part) {
            gFocusedProjectView.manager.addItem(item.part.clone(), parentindex);
        }
    }        
}

function OpenDirectoryShortcut(part) {
    var defaultDirectory = part.value;
    var paths = filepicker_openFiles(defaultDirectory);
    if (paths == null)
        return;
    open_openMultipleURIs(paths);
}

function peFile_CompareTwoFiles()
{
    // We've been picked from the tools menu
    fname = filepicker_openFile(null, null, "Select first file");
    if (!fname) return;
    otherfile = filepicker_openFile(null, null, "Select second file");
    if (!otherfile) return;
    peFile_ShowDiffs(fname, otherfile);
}

function peFile_ShowDiffs(fname1, fname2) {
    // fname1/2 can be filenames or uri's.
    fname1 = fname1.replace("\\", "\\\\", 'g');
    fname2 = fname2.replace("\\", "\\\\", 'g');
    // difflib can be slow, so we do this in a timeout
    window.setTimeout("_openDiffWindowForFiles('"+fname1+"', '"+fname2+"');",
                      0);
}

