# Copyright (c) 2004 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

# Utility functions for dealing with scimoz & indentation
# that are used both by koLanguageServiceBase and koLanguageCommandHandler

def makeIndentFromWidth(scimoz, width):
    """ returns a string consisting of however many tabs and spaces
    are needed to make a width of `width`, using the useTabs pref
    and the scimoz tabWidth
    """
    if scimoz.useTabs:
        tabWidth = scimoz.tabWidth
        # guard against a misconfigured scimoz with a tabWidth
        # of 0, which would cause a divide by zero error
        if tabWidth == 0: tabWidth = 8
        numtabs, numspaces = divmod(width, scimoz.tabWidth)
        return '\t'*numtabs + ' '*numspaces
    else:
        return ' '*width

def adjustClosingXMLTag(scimoz):
    """ This function is called when a ">" from an XML end tag
    is inserted, and it will shift the current line either back
    or forwards so that the tag is aligned with the matching start tag.
    It will only do so if the end tag is the only thing to the left of
    the current position on the current line.
    """
    charPos = scimoz.currentPos-1
    # find out what tag we just closed
    beforeText = scimoz.getTextRange(0, scimoz.currentPos)
    # (for now, assuming no intervening space
    leftCloseIndex = beforeText.rfind('</')
    if leftCloseIndex == -1:
        # no idea what to do in this case, just bail
        return
    tagName = beforeText[leftCloseIndex+len('</'):scimoz.currentPos-1].strip()
    # again, ignore whitespace between < and tagname
    startTag = beforeText[:leftCloseIndex-1].rfind('<' + tagName)
    startIndentWidth = scimoz.getColumn(startTag)
    indent = scimoz.indent
    if indent == 0:
        indent = 8
    indentLevel, extras = divmod(startIndentWidth, indent)
    if indentLevel and not extras:
        indentLevel -= 1
    nextIndentWidth = indentLevel * scimoz.indent
    lineNo = scimoz.lineFromPosition(charPos-1)
    startOfLine = max(0, scimoz.positionFromLine(lineNo))
    if startOfLine > leftCloseIndex:
        return
    stuffToLeft = scimoz.getTextRange(startOfLine, leftCloseIndex)
    if not stuffToLeft.strip(): # XXX Do we want to pref this?
        # we can align the comment closing before we do the newline
        indent = makeIndentFromWidth(scimoz, scimoz.getColumn(startTag))
        scimoz.targetStart = scimoz.positionFromLine(lineNo)
        scimoz.targetEnd = leftCloseIndex
        scimoz.replaceTarget(len(indent), indent)
    scimoz.chooseCaretX()
